--[[ Script Info:

    • Script:       Shield Storage
    • By:           Labyrinthia Rosenfeld (mb.srb2 profile: https://mb.srb2.org/members/labyrinthia.6409/)
    • Last edit:    12/01/2023 (dd/MM/yyyy)
    • Version:      2.0
    • Description:  Allows the player to store, summon, and exchange shields so you don't have to stick with
                    just one shield for an entire stage. The cost of summoning/exchanging can be altered
                    and it has pseudo-support for custom shields (you'll have do edit this script).

    • DISCLAIMER:   This script depends on l_ButtonState.lua for button presses validation.
                    The aforementioned script will come along with this one but l_ButtonState.lua is in NO WAY, SHAPE, OR FORM my script.
                    It was taken from a GitHub repository with a multitude of (allegedly) free to use scripts "for SRB2 modders to
                    use in their projects." (SOURCE: https://github.com/sprkizard/srb2-lua-lib)
                    The l_ButtonState.lua that was included in the pk3 archive containing this script is UNCHANGED, 
                    and credits for the script goes to their respective owners.
                
    • PS:           If you're wondering why the hell I'm leaving such massive comments on a script most people will probably 
                    not even bother to read, it's actually mostly so I can remember what I did when I go fiddle with this again.
                    Like, really, the original release was in 2018 and as of the time of this comment, it's 2023. 
                    I have no idea how the original script works aside from faint memories of me suffering to get x and y things to work, 
                    and in all honesty, I'm surprised it worked at all looking at it now, lmao.]]

-- TODO: Better graphics for the selector. I know I am no artist but GOSH that thing is UGLY.

-- Change this if you think 5 rings is too low/high of a cost to exchange/summon shields.
local RING_COST = 5

local lookupTable
--[[ shieldLookup() explanation: 

    Since I decided to be a butt and have the possibility to include "fancy" names and descriptions to each shield, I'd have to either
    set them up individually each time I checked the shield decimal value (I'm not using the constants, sorry), 
    which I honestly hate thinking about. So instead I decided to use a lookup table to do the same thing but in an arguably 
    more concise way, so if I wanna add more useless flair to this later, it won't be THAT much of a pain. 
    
    Me from a literal day later:
    I'm so glad I did it this way, this made working with the TV boxes icons and sfx for each shield so much easier.]]
local function shieldLookup(decimalValue)
    --[[ ~READ~ MAKING THIS WORK WITH CUSTOM SHIELDS:
        Simply add another key — value pair in the lookupTable below following the respective format:
        [DECIMAL_VALUE_OF_THE_SHIELD_AS_STRING] = {SHIELD_NAME, DECIMAL_VALUE, DESCRIPTION, FILENAME_OF_THE_ICON, SFX_TO_PLAY_ON_SUMMON}

        Don't forget to add the comma in the most recent entry beforehand (It happened to me a lot of times, 
        trust me I'm not messing with you).
        With all of that done, it should (hopefully) work.]]
    lookupTable = lookupTable or {
        ['0']    = {'No Shield',         0,    'No description.', 'NO ICON',  sfx_None  },
        ['1']    = {'Pity Shield',       1,    'No description.', 'TVPIICON', sfx_shield},
        ['2']    = {'Whirlwind Shield',  2,    'No description.', 'TVWWICON', sfx_wirlsg},
        ['3']    = {'Amargeddon Shield', 3,    'No description.', 'TVARICON', sfx_armasg},
        ['4']    = {'Pink Shield',       4,    'No description.', 'TVPPICON', sfx_shield},
        ['3073'] = {'Elemental Shield',  3073, 'No description.', 'TVELICON', sfx_elemsg},
        ['4097'] = {'Attraction Shield', 4097, 'No description.', 'TVATICON', sfx_attrsg},
        ['1025'] = {'Flame Shield',      1025, 'No description.', 'TVFLICON', sfx_s3k3e },
        ['2049'] = {'Bubble Shield',     2049, 'No description.', 'TVBBICON', sfx_s3k3f },
        ['4098'] = {'Lightning Shield',  4098, 'No description.', 'TVZPICON', sfx_s3k41 },
        ['256']  = {'Force Shield',      256,  'No description.', 'TVFOICON', sfx_frcssg}
    }

    --[[ this goddamn wizardry explanation: 
        
        You might be wondering: "What the HELL is going on here?" Well, this is all because of Force Shield.
        Simply because the 256 decimal value is equivalent to Force Shield with only 1 HP left, and its 2 HP "variant" has a 
        decimal value of 257, following that logic, decimal values from 258 onwards (from my experimentation), will add another
        hitpoint to the Force Shield until it hits 1025, which is the decimal value for the Flame Shield (it probably actually has an
        upper limit lower than that but you're crazy if you think I'll be checking values 258 — 1024 ONE BY ONE to confirm my speculations).
        
        So the way that I work around that is simply assuming that any decimal value that isn't on the table will be just another Force Shield with
        some bigass HP value, so when the game tries to index the lookupTable with it, it'll return nil, this default value will be assigned to it instead.]]
    local shieldInfo = lookupTable[tostring(decimalValue)] or {'Force Shield', decimalValue, 'No description.', 'TVFOICON', sfx_forcsg}
    return shieldInfo
end

--[[ ShieldObj... object:
    I'm mostly updating this old thing because I've been reminiscing about old things and I have a small interest in playing
    around with object oriented programming in Lua. The old script only worked the way it did because I had no idea how to
    actually store shields in a similar way to how it works with Mighty in Sonic Megamix (which was the inspiration for this addon),
    and I'm going to change that this time by applying the power of object orientation and a semi-lethal dosage of caffeine and breakcore.
    ...and cursing at my screen when something doesn't work.]]
local ShieldObj = {
    name = 'Shieldless? :megamind:',
    decimalValue = 0,
    desc = 'L + Ratio + No emeralds + You fell off + Touch greenflower zone + That\'s why Greenflower City was destroyed',
    icon = 'SOMETHING_ICON',
    sfx = sfx_None
}
function ShieldObj:new(obj, shieldDecimal)
    obj = obj or {}
    obj.name, obj.decimalValue, 
    obj.desc, obj.icon, 
    obj.sfx = unpack(shieldLookup(shieldDecimal))
    
    setmetatable(obj, ShieldObj)
    self.__index = self
    return obj
end

-- Hud Function
local function shieldDisplay(v, stplyr)
    if stplyr and stplyr.valid and stplyr.mo and stplyr.mo.valid then
        local Scale = FRACUNIT/2
        local Snapped = V_SNAPTOTOP
        local InitialXDist = 300
		local InitialYDist = 1--13
		local DistRateX = 0
		local DistRateY = 18

        for i, j in ipairs(stplyr.shieldstorage) do
            v.drawScaled((InitialXDist + (DistRateX * i))*FRACUNIT, (InitialYDist + (DistRateY * i))*FRACUNIT, (Scale), v.cachePatch(j.icon),V_SNAPTORIGHT|Snapped)
        end

        -- Don't draw the selector to the hud if you have less than 2 shields.
        if #stplyr.shieldstorage < 2 then 
            return 
        else
            v.drawScaled((InitialXDist + (DistRateX * stplyr.shieldstorage_index))*FRACUNIT, (InitialYDist + (DistRateY * stplyr.shieldstorage_index))*FRACUNIT, (Scale), v.cachePatch("SLCTFM"),V_SNAPTORIGHT|Snapped)
        end
    end
end
hud.add(shieldDisplay, "game")

local function shieldWorker(player)
    --[[ player properties: 

        Create the necessary player properties. In the original script I was a dumbass and for some reason thought
        having a bajillion variables in the outer scope for everything was a smart idea, but that just meant that it
        couldn't be used for multiplayer gamemodes (on top of looking horrible). I'm still not sure if I'm gonna bother 
        making this netplay compatible but at least I wanna make things the right way.]]
    if not player.shieldstorage then player.shieldstorage = {} end
    if not player.shieldstorage_index then player.shieldstorage_index = 1 end
    if (player.pflags & PF_FINISHED) or 
       (player.pflags & PF_GAMETYPEOVER) then 
        player.shieldstorage = {}
        player.shieldstorage_index = 1
        return
    end

    --[[ checking button presses: 

        I'll be using l_ButtonState.lua to validate button presses because, in all honesty, this is way better than anything
        I could ever come up with and I simply cannot be bothered.
        
        On a side-note: 
        I got this (script is way down there) from https://github.com/sprkizard/srb2-lua-lib/blob/master/player/l_ButtonState.lua 
        and I have no idea why it's in a github repo instead of the forums, I mean I appreciate the idea but, is there
        any specific reason why???]]
    if player.buttonstate[BT_WEAPONPREV] == 1 then
        if player.shieldstorage_index >= #player.shieldstorage then
            player.shieldstorage_index = 1
        else
            player.shieldstorage_index = $1 + 1
        end
        if #player.shieldstorage >= 2 then
            S_StartSound(nil, sfx_wepchg, player)
        end
    elseif player.buttonstate[BT_WEAPONNEXT] == 1 then
        if player.shieldstorage_index <= 1 then
            player.shieldstorage_index = max(1, #player.shieldstorage)
        else
            player.shieldstorage_index = $1 - 1
        end
        if #player.shieldstorage >= 2 then
            S_StartSound(nil, sfx_wepchg, player)
        end
    end

    if player.buttonstate[BT_CUSTOM1] == 1 then
        A_StoreShield(player.mo, player.powers[pw_shield], nil)
    end
    if player.buttonstate[BT_CUSTOM2] == 1 then
        if player.rings >= RING_COST then
            if player.shieldstorage[player.shieldstorage_index] then
                local shieldObj = player.shieldstorage[player.shieldstorage_index]
                table.remove(player.shieldstorage, player.shieldstorage_index)
                if player.shieldstorage_index > #player.shieldstorage then
                    player.shieldstorage_index = #player.shieldstorage
                end
                player.mo.target = player.mo
                A_SummonShield(player.mo, shieldObj.decimalValue, shieldObj.sfx)
                player.rings = $1 - RING_COST
            end
        end
    end
end
--[[ PlayerThink hook:
    
    Since when is PlayerThink a hook??? Gosh, I've been away from this game for a long time. 
    Anyways, add the main shield function into the hook.]]
addHook("PlayerThink", shieldWorker)

-- function to clear the player's stored shields on death (the other function probably handles it already but better safe than sorry).
local function begoneShield(target, inflictor, source, damagetype)
    target.player.shieldstorage = {}
    target.player.shieldstorage_index = 1
    return false
end
addHook("MobjDeath", begoneShield, MT_PLAYER)

-- function to clear the player's stored shields on spawn (for preventing carrying shields over death/retries/map changes).
local function resetShield(player)
    player.shieldstorage = {}
    player.shieldstorage_index = 1
end
addHook("PlayerSpawn", resetShield)

-- creating new action to store shields because I can (actually it's because it's more useful for me that way).
function A_StoreShield(actor, var1, var2)
    local shieldType = var1
    local player = actor.player

    if not actor.player then return end
    if player.powers[pw_shield] == SH_NONE then return end

    table.insert(player.shieldstorage, ShieldObj:new(nil, shieldType))
    if #player.shieldstorage > 5 then
        table.remove(player.shieldstorage, 5)
    end
    if #player.shieldstorage == 1 then
        player.shieldstorage_index = max(#player.shieldstorage, player.shieldstorage_index)
    end
    player.powers[pw_shield] = SH_NONE
    P_SpawnShieldOrb(player)
    --chchng was too obnoxious so I removed the sound from when you store a shield. Gonna stay that way until I find a replacement.
    --S_StartSound(nil, sfx_chchng, player)
end

--[[ manual recreation of A_GiveShield override:
    WHY CAN'T YOU CALL SUPER() OUTSIDE OF A STATE CHANGE??? 
    I NEED TO MAKE A ~WHOLE~ NEW ACTION FOR THIS INSTEAD OF JUST MANUALLY CALLING A_GiveShield UGH.]]
function A_SummonShield(actor, var1, var2)
    if not actor.target or not actor.target.player then return end
    local player = actor.target.player
    local locvar1 = var1
    local locvar2 = var2
    A_StoreShield(player.mo, player.powers[pw_shield], nil)

    P_SwitchShield(player, locvar1);
	S_StartSound(player.mo, locvar2)
end

-- A_GiveShield override so shield storing happens when breaking shield monitors.
function A_GiveShield(actor, var1, var2)
    if not actor.target or not actor.target.player then return end
    local player = actor.target.player
    A_StoreShield(player.mo, player.powers[pw_shield], nil)
    super(actor, var1, var2)
end

--[[ Post-script message:
    This is more of an emotional outlet more than anything so feel free to ignore this if you want.
    It's been a little more than 4 years since I released Shield Storage initially for 2.1.X, and 5 years since I began
    fiddling with LUA in SRB2 at all (god time really does fly, huh?)

    I never thought that programming would be more than a silly side hobby that I'd eventually drop like the ones I had before,
    honestly, I didn't think I had it in me. But here I am, 5 years into the future, with a fullstack and software development job.

    And as silly as it sounds, it all started here. From Lua I got into other languages, got into college, and eventually landed a pretty
    comfy job. I always struggled in deciding a career for myself, so I can't help but to imagine that I'd have been lost if not for
    following a programming career.

    So for this reason, thank you. Thank you Sonic Team Jr. for keeping this project alive for so long,
    thank you the SRB2 community for never losing interest in the game, and thank you to all the other
    scripters (modders? whatever the right term may be) for inspiring the curiosity and creativity that I
    thought I never had. Seriously.

    For anyone reading this and finding yourself in a similar situation that I was 4~5 years ago:
    Don't give up. Whatever your aspirations are, you can do it. With enough perseverance and discipline surely
    you can achieve your goals. So keep going.
]]